﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Threading.Tasks;
using System.Xml.Linq;
using Mobile.Search.Shared;
using Mobile.Search.Shared.Models;
using ServiceStack;
using System.Text.RegularExpressions;

namespace Mobile.Search.Providers
{
    public class TaboolaProvider
    {
        private static readonly List<string> UserAgents = new List<string>
        {
            "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2272.101 Safari/537.36",
            "Mozilla/5.0 (Linux; U; Android 4.0.3; ko-kr; LG-L160L Build/IML74K) AppleWebkit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30"
        };

        public static SearchResult GetListings(string ip, int count, string sourceUrl, string userReferrer, string utmCampaign)
        {

            // Example: http://localhost:64292/search/template?templateKey=dynamic_css_taboola&keyword=best+buy+tv&debug=true&content_subtype=WeatherHeader&storeClickUrls=true&emailId=f1ff24b475da4d30af21925bc9562fb6&Ip=73.157.247.196
            var userAgent = UserAgents[1];
            
            // Make sure incoming IP is a valid one!
            Match match = Regex.Match(ip, @"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}");
            if (!match.Success)
            {
                ip = "";
            }

            var url = MakeUrl(count, sourceUrl, userReferrer, utmCampaign, ip);

            var rawTaboolaJsonData = GetJson(url, userAgent);

            var listings = GetListingsFromJson(rawTaboolaJsonData).ToList();

            return new SearchResult
            {
                SearchId = Guid.NewGuid(),
                Ip = ip,
                UserAgent = userAgent,
                SourceTag = null,
                TypeTag = null,
                Listings = listings,
                ListingsCount = listings.Count(),
                sourceUrl = sourceUrl,
                userReferrer = userReferrer,
                utmCampaign = utmCampaign
            };
        }

        private static string MakeUrl(int count, string sourceUrl, string userReferrer, string utmCampaign, string ip)
        {
            var url = "http://taboola-api.rm-mobile.com/api/taboola/RecommendationsGet"
                .AddQueryParam("appType", "mobile")
                .AddQueryParam("placementRecCount", count)      // Ex: 5
                .AddQueryParam("placementOrganicType", "video")
                .AddQueryParam("userSession", "init")
                .AddQueryParam("sourceType", "search")
                .AddQueryParam("sourceId", "214321562187")
                .AddQueryParam("sourceUrl", sourceUrl)          // Ex: www.google.com
                .AddQueryParam("userReferrer", userReferrer)    // Ex: Coupons
                .AddQueryParam("utmCampaign", utmCampaign)      // Ex: Coupons
                .AddQueryParam("userRealIp", ip);               // Ex: 1.2.3.4

            return url;
        }

        private static string GetJson(string url, string userAgent)
        {
            try
            {

                return url.GetStringFromUrl(
                    requestFilter: req =>
                    {
                        req.Method = "GET";
                        req.KeepAlive = false;
                        req.UserAgent = userAgent;
                        req.Headers.Add("Accept-Encoding", "gzip,deflate");
                        req.Headers.Add("Cache-Control", "no-cache");
                    });
            }
            catch (WebException webEx)
            {
                //TODO: Log
                if (webEx.IsBadRequest())
                {
                    throw new Exception(webEx.GetResponseBody(), webEx);
                }
                throw;
            }
            //return "";
        }

        public static IEnumerable<Listing> GetListingsFromJson(string json)
        {
            var rank = 1;
            var taboolaRequest = json.FromJson<TaboolaResponse>();
            return taboolaRequest.list.Select(x => new Listing
            {
                Rank = rank++,
                Title = x.name,
                Description = x.description,
                BiddedListing = true,
                ResultType = default(string),
                SiteHost = x.branding,
                Categories = x.categories,
                SiteLanguage = default(string),
                ListingLanguage = default(string),
                PhoneNumber = default(string),
                ClickUrl = x.url,
                ThumbnailImageUrl = x.thumbnail[0].url
            });
        }

        private static string TryGetSiteHostFromURL(XElement element, string name)
        {
            var host = default(string);
            var url = element.Element("url").Value;
            if (url.Length > 0)
            {
                Uri uri = new Uri(url);
                host = uri.Host;
            }
            return host;
        }

        private static string TryGetElementValue(XElement element, string name)
        {
            var e = element.Element(name);
            return e != null ? e.Value : default(string);
        }
    }

    public class TaboolaResponse
    {
        public string session { get; set; }
        public string id { get; set; }
        public string ui { get; set; }
        public List<TaboolaResponseItem> list { get; set; }
    }

    public class TaboolaResponseItem
    {
        public List<TaboolaThumbnailUrl> thumbnail { get; set; }
        public string description { get; set; }
        public string type { get; set; }
        public string name { get; set; }
        public string created { get; set; }
        public string branding { get; set; }
        public string duration { get; set; }
        public List<String> categories { get; set; }
        public string views { get; set; }
        public string itemId { get; set; }
        public string origin { get; set; }
        public string url { get; set; }
    }

    public class TaboolaThumbnailUrl
    {
        public string url { get; set; }
    }
}
