﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Threading.Tasks;
using System.Xml.Linq;
using Mobile.Search.Shared;
using Mobile.Search.Shared.Models;
using Serilog;
using ServiceStack;
using System.Diagnostics;

namespace Mobile.Search.Providers
{ 
    public class YahooProvider
    {
      
        private static readonly List<string> UserAgents = new List<string>
        {
            "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2272.101 Safari/537.36",
            "Mozilla/5.0 (iPhone; CPU iPhone OS 9_2_1 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13D15 Safari/601.1"
        };


        public static async Task<IEnumerable<Listing>> GetListingsAsync(string query,int sourceTagId, string typeTag, string serveUrl,int biddedResults = 5, int nonBiddedResults = 8,string userAgent = null, string ip = null)
        {
            var sourceTag = YahooTags.GetSourceTag(sourceTagId);

            userAgent = userAgent ?? UserAgents[0];
            ip = ip ?? "";
            var url = MakeUrl(ip, userAgent,query,typeTag, serveUrl,biddedResults, nonBiddedResults,sourceTag: sourceTag);

            return GetListingsFromXml(await GetXmlAsync(url, userAgent));
        }

        public static List<string> GetRelatedKeywords(string keyword)
        {
            var searchRequest = new SearchRequest
            {
                Ip = "205.232.172.74",
                Query = keyword,
                ServeUrl = "http://mobsearch.adk-mobile.com/search",
                TypeTag = "get_related_searches",
                UseFakeUserAgent = true,
                BiddedResults = 1,
                NonBiddedResults=0,
                Timestamp = DateTime.UtcNow,
                SourceTag = "adknowledge_2click_search_b"
            };

            return GetListings(searchRequest).RelatedSearches;
        }

        public static SearchResult GetListings(SearchRequest request)
        {
            var sourceTagId = request.SourceTagId != 0 ? request.SourceTagId : 1;
            var sourceTag = !string.IsNullOrEmpty(request.SourceTag) ? request.SourceTag : YahooTags.GetSourceTag(sourceTagId);

            var userAgent = request.UserAgent;
            if (request.UseFakeUserAgent || string.IsNullOrEmpty(request.UserAgent)) // we could have android come through ||  !request.UserAgent.ToLower().ContainsAny("iphone", "ipad", "ipod"))
            {
                userAgent = UserAgents[1];
            }
            request.Ip = request.Ip ?? "::1";

            // Store the original query in case we end up changing it
            var originalQuery = request.Query;

            var url = MakeUrl(request.Ip,userAgent, request.Query, request.TypeTag, request.ServeUrl,
                request.BiddedResults, request.NonBiddedResults, sourceTag: sourceTag);

            var xml = GetXml(url, request.UserAgent);

            var listings = GetListingsFromXml(xml)
                .ToList();

            var suggestions = GetSuggestedResultsFromXml(xml);

            // we returned no listings, and the query contains multiple words
            if (!listings.Any() && request.Query.ContainsAny(" ", "+"))
            {
                // try to query again with one less word psychic+readings+online -> psychic+readings 
                request.Query = request.Query.Remove(request.Query.LastIndexOfAny(new[] {' ', '+'}));
                url = MakeUrl(request.Ip,userAgent, request.Query, request.TypeTag, request.ServeUrl,
                    request.BiddedResults, request.NonBiddedResults, sourceTag: sourceTag);
                xml = GetXml(url, request.UserAgent);
                listings = GetListingsFromXml(xml)
                    .ToList();
            }

            return new SearchResult
            {
                SearchId = Guid.NewGuid(),
                Ip = request.Ip,
                UserAgent = request.UserAgent,
                SourceTag = sourceTag,
                TypeTag = request.TypeTag,
                Listings = listings,
                ListingsCount = listings.Count,
                Query = request.Query,
                RelatedSearches = suggestions,
                OriginalQuery = originalQuery
            };
        }
        public static SearchResult GetListings(string query,int sourceTagId, string typeTag, string serveUrl, int biddedResults = 5, int nonBiddedResults = 8,
            string userAgent = null, string ip = null)
        {
            var sourceTag = YahooTags.GetSourceTag(sourceTagId);
            userAgent = userAgent ?? UserAgents[1];
            ip = ip ?? "";
            var url = MakeUrl(ip, userAgent,query,typeTag, serveUrl,biddedResults, nonBiddedResults);
            var xml = GetXml(url, userAgent);

            //will wait for task to be completed
            var nextTuple = GetNextArgsFromXml(xml);
            var listings = GetListingsFromXml(xml);
            return new SearchResult
            {
                SearchId = Guid.NewGuid(),
                Ip = ip,
                UserAgent = userAgent,
                SourceTag = sourceTag,
                TypeTag = typeTag,
                Listings = listings,
                ListingsCount = listings.Count(),
                Query = query,
                NextUrl = nextTuple != null
                    ? MakeUrl(ip, userAgent, query,typeTag,serveUrl, biddedResults, nonBiddedResults, nextTuple.Item1, nextTuple.Item2)
                    : "",
            };
        }

        private static string MakeUrl(string ip, string userAgent, string query, string typeTag,string serveUrl, int biddedResults, int nonBiddedResults, string xargs = null, string hData = null, string sourceTag = null)
        {
            sourceTag = sourceTag ?? YahooTags.GetSourceTag(1);
            var url = "https://xml-nar-ss.ysm.yahoo.com/d/search/p/adknowledge/xml/"
                .AddQueryParam("Partner", sourceTag)
                .AddQueryParam("config", 1234567890)
                .AddQueryParam("serveUrl", serveUrl)
                .AddQueryParam("affilData", "ip={0}&ua={1}".Fmt(ip, userAgent))
                .AddQueryParam("adultFilter", "clean")
                .AddQueryParam("Keywords", query)
                .AddQueryParam("mkt", "us")
                .AddQueryParam("on", biddedResults)
                .AddQueryParam("in", nonBiddedResults)
                .AddQueryParam("longAdsTitle", "1")
                .AddQueryParam("osb", "1")
                .AddQueryParam("adEnableMerchRating", "1")
                .AddQueryParam("enableFavicon", "1")
                .AddQueryParam("moreSponsoredResults", "1")
                .AddQueryParam("enableImageInAds", "1")
                .AddQueryParam("type", typeTag)

                // New Yahoo! API Extensions 2017-10-11
                .AddQueryParam("adEnableCalloutExtension", "1") // 1. Callout Extension             (Call to Action)
                .AddQueryParam("adEnableReview", "1")           // 2. Review Extension              (Review from Official 3rd Party)
                .AddQueryParam("adEnable4thLine", "1")          // 3. 4th Line Extension            (Added metadata)
                .AddQueryParam("adEnableTopAd", "1")            // 4. Top Ad Extension              (Popularity / clicks on site)
                .AddQueryParam("adEnableSecurityBadge", "1")    // 5. Security Extension            (Secured by Official Security Provider)
                .AddQueryParam("adEnableBrand", "1")            // 6. Brand Extension               (Website by Official Brand)
                .AddQueryParam("adSmartAnnotations", "1");      // 7. Smart Annotation Extension    (Short, Sexy Site Descriptor)

            //.AddQueryParam("c2c", "1")
            //.AddQueryParam("siteLink", "1");

            if (xargs != null)
            {
                url.AddQueryParam("xargs", xargs, false);
            }
            if (hData != null)
            {
                url.AddQueryParam("hData", hData, false);
            }
            
            // System.Diagnostics.Process.Start(url);
            // Debug.WriteLine(url);
            return url;
        }

        private static async Task<string> GetXmlAsync(string url, string userAgent)
        {
            return await url.GetStringFromUrlAsync(
                requestFilter: req =>
                {
                    req.Method = "GET";
                    //req.KeepAlive = false;
                    req.UserAgent = userAgent;
                    req.Headers.Add("Accept-Encoding", "gzip,deflate");
                    req.Headers.Add("Cache-Control", "no-cache");
                });
        }

        private static string GetXml(string url, string userAgent, int retries = 2)
        {
            while (true)
            {
                try
                {

                    return url.GetStringFromUrl(
                        requestFilter: req =>
                        {
                            req.Method = "GET";
                            //req.KeepAlive = false;
                            req.UserAgent = userAgent;
                            req.Headers.Add("Accept-Encoding", "gzip,deflate");
                            req.Headers.Add("Cache-Control", "no-cache");
                        });
                }
                catch (WebException webEx)
                {
                    if (retries > 0)
                    {
                        Log.Error(webEx, "Error Getting XML from Yahoo at {url} (retrying)", url);
                        retries--;
                    }
                    else
                    {
                        Log.Error(webEx, "Error Getting XML from Yahoo at {url} (re-throwing exception)", url);
                        if (webEx.IsBadRequest())
                        {
                            throw new Exception(webEx.GetResponseBody(), webEx);
                        }
                        throw;
                    }
                }
            }
        }

        /// <summary>
        /// Extracts the xargs, and hData from the NextArgs xml node
        /// </summary>
        /// <returns>Tuple of xargs, hdata</returns>
        private static Tuple<string,string> GetNextArgsFromXml(string xml)
        {
            var xmlDoc = XDocument.Parse(xml).Root;
            if (xmlDoc == null) return null;
            var nextArgs = xmlDoc.Element("NextArgs");
            if (nextArgs == null) return null;

            var val = nextArgs.Value;
            var startOfXargs = val.IndexOf("xargs=") + 6;
            var startOfHdata = val.IndexOf("hData=") + 6;

            var xargs = val.Substring(startOfXargs, val.IndexOf("&", startOfXargs) + 1);
            var hData = val.Substring(startOfHdata);
            return Tuple.Create(xargs, hData);
        }

        public static List<string> GetSuggestedResultsFromXml(string xml)
        {
            var xmlDoc = XDocument.Parse(xml).Root;
            if (xmlDoc != null)
            {
                return xmlDoc.Elements("MoreSponsoredResult")
                    .SelectMany(x => x.Elements("Suggestion"))
                    .SelectMany(x => x.Elements("name"))
                    .Select(x => x.Value)
                    .ToList();
            }
            return new List<string>();
        }

        public static IEnumerable<Listing> GetListingsFromXml(string xml)
        {
            var xmlDoc = XDocument.Parse(xml).Root;
            if (xmlDoc != null)
            {
                return xmlDoc.Elements("ResultSet")
                    .SelectMany(x => x.Elements("Listing"))
                    .Select(x => new Listing
                    {
                        Rank = int.Parse(TryGetAttribute(x, "rank") ?? "1"),
                        Title = StripHtmlTags(TryGetAttribute(x, "title")),
                        Description = StripHtmlTags(TryGetAttribute(x, "description")),
                        SiteHost = TryGetAttribute(x, "siteHost"),
                        BiddedListing = bool.Parse(TryGetAttribute(x, "biddedListing") ?? "true"),
                        ResultType = TryGetAttribute(x, "resultType"),
                        SiteLanguage = TryGetAttribute(x, "siteLang"),
                        ListingLanguage = TryGetAttribute(x, "listingLang"),
                        PhoneNumber = TryGetAttribute(x, "phoneNumber"),
                        ClickUrl = x.Element("ClickUrl") != null ? x.Element("ClickUrl").Value : "",
                        FaviconUrl = TryGetAttribute(x, "faviconUrl"),
                        MerchRatingNumber = TryGetAttribute(x, "MRRating"),
                        MerchRatingUrl = TryGetAttribute(x, "MRStarLink"),
                        //LongTitle = StripHtmlTags(TryGetAttribute(x, "longAdsTitle")),
                        ImageUrl = TryGetAttribute(x, "imageInAdsUrl_V1"),
                        SiteLinks = TryGetSiteLinks(x),

                        // === Yahoo! API Extensions Update 2017-10-11 === //

                        // 1. Callout Extensions
                        CalloutAnnotation = TryGetCalloutExtension(x),

                        // 2. Review Extensions
                        ReviewSource = TryGetReviewSource(x),
                        ReviewText = TryGetReviewText(x),

                        // 3. 4th Line Annotations
                        Ad4thLine = TryGetAd4thLine(x),

                        // 4. Top Ad Annotations
                        HasTopAdAnnotation = TryGetHasTopAdAnnotation(x),
                        TopAdDomain = TryGetTopAdDomain(x),
                        TopAdClicks = TryGetTopAdClicks(x),

                        // 5. Security Badge
                        SecurityProvider = TryGetSecurityProvider(x),

                        // 6. Brand Annotations
                        AdOfficialBrand = TryGetAdOfficialBrand(x),

                        // 7. Smart Annotations
                        SmartAdAnnotation = TryGetSmartAdAnnotation(x)
                    });
            }
            return new List<Listing>();
        }

        private static string TryGetAttribute(XElement element, string name)
        {
            var val = default(string);
            try
            {
                val = element.Attribute(name).Value;
            }
            catch
            {
                val = default(string);
            }
            return val;
        }

        // http://www.dotnetperls.com/remove-html-tags
        public static string StripHtmlTags(string source)
        {
            char[] array = new char[source.Length];
            int arrayIndex = 0;
            bool inside = false;

            for (int i = 0; i < source.Length; i++)
            {
                char let = source[i];
                if (let == '<')
                {
                    inside = true;
                    continue;
                }
                if (let == '>')
                {
                    inside = false;
                    continue;
                }
                if (!inside)
                {
                    array[arrayIndex] = let;
                    arrayIndex++;
                }
            }
            return new string(array, 0, arrayIndex);
        }

        private static List<Listing.SiteLink> TryGetSiteLinks(XElement element)
        {
            List<Listing.SiteLink> siteLinks;
            if (element.Element("Extensions") != null && element.Element("Extensions").Elements("SiteLinks") != null)
            {
                siteLinks = element.Element("Extensions")
                    .Elements("SiteLinks")
                    .SelectMany(l => l.Elements("SiteLink"))
                    .Select(l => new Listing.SiteLink
                    {
                        Text = l.Element("Text").Value,
                        Url = l.Element("Url").Element("Url").Value
                    }).ToList();
            }
            else
            {
                siteLinks = new List<Listing.SiteLink>();
            }
            return siteLinks;
        }

        // === Yahoo! API Extensions Update 2017-10-11 === //

        // 1. Callout Extensions
        // CalloutAnnotation = x.Element("Extensions").Element("calloutExtension").Element("phrases").Element("phrase").Value ?? "",

        private static String TryGetCalloutExtension(XElement element)
        {
            String calloutExtension;
            if (element.Element("Extensions") != null 
                && element.Element("Extensions").Element("calloutExtension") != null
                && element.Element("Extensions").Element("calloutExtension").Element("phrases") != null
                && element.Element("Extensions").Element("calloutExtension").Element("phrases").Element("phrase") != null)
            {
                calloutExtension = element.Element("Extensions").Element("calloutExtension").Element("phrases").Element("phrase").Value;
            }
            else
            {
                calloutExtension = "";
            }
            return calloutExtension;
        }

        // 2. Review Extensions
        // ReviewSource = x.Element("Extensions").Element("reviewExtension").Element("sourceText").Value ?? "",
        private static String TryGetReviewSource(XElement element)
        {
            String ReviewSource;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("reviewExtension") != null
                && element.Element("Extensions").Element("reviewExtension").Element("sourceText") != null)
            {
                ReviewSource = element.Element("Extensions").Element("reviewExtension").Element("sourceText").Value;
            }
            else
            {
                ReviewSource = "";
            }
            return ReviewSource;
        }
        // ReviewText = x.Element("Extensions").Element("reviewExtension").Element("text").Value ?? "",
        private static String TryGetReviewText(XElement element)
        {
            String ReviewText;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("reviewExtension") != null
                && element.Element("Extensions").Element("reviewExtension").Element("text") != null)
            {
                ReviewText = element.Element("Extensions").Element("reviewExtension").Element("text").Value;
            }
            else
            {
                ReviewText = "";
            }
            return ReviewText;
        }

        // 3. 4th Line Annotations
        // Ad4thLine = x.Element("Extensions").Element("Ad4thLineAnnotation").Element("textElements").Element("textElement").Element("text").Value ?? "",
        private static String TryGetAd4thLine(XElement element)
        {
            String Ad4thLine;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("Ad4thLineAnnotation") != null
                && element.Element("Extensions").Element("Ad4thLineAnnotation").Element("textElements") != null
                && element.Element("Extensions").Element("Ad4thLineAnnotation").Element("textElements").Element("textElement") != null
                && element.Element("Extensions").Element("Ad4thLineAnnotation").Element("textElements").Element("textElement").Element("text") != null)
            {
                Ad4thLine = element.Element("Extensions").Element("Ad4thLineAnnotation").Element("textElements").Element("textElement").Element("text").Value;
            }
            else
            {
                Ad4thLine = "";
            }
            return Ad4thLine;
        }

        // 4. Top Ad Annotations
        // HasTopAdAnnotation = x.Element("Extensions").Element("topAdExtension").Element("isTopAd").Value ?? "",
        private static String TryGetHasTopAdAnnotation(XElement element)
        {
            String HasTopAdAnnotation;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("topAdExtension") != null
                && element.Element("Extensions").Element("topAdExtension").Element("isTopAd") != null)
            {
                HasTopAdAnnotation = element.Element("Extensions").Element("topAdExtension").Element("isTopAd").Value;
            }
            else
            {
                HasTopAdAnnotation = "";
            }
            return HasTopAdAnnotation;
        }
        // TopAdDomain = x.Element("Extensions").Element("topAdExtension").Element("adDomainName").Value ?? "",
        private static String TryGetTopAdDomain(XElement element)
        {
            String TopAdDomain;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("topAdExtension") != null
                && element.Element("Extensions").Element("topAdExtension").Element("adDomainName") != null)
            {
                TopAdDomain = element.Element("Extensions").Element("topAdExtension").Element("adDomainName").Value;
            }
            else
            {
                TopAdDomain = "";
            }
            return TopAdDomain;
        }
        // TopAdClicks = x.Element("Extensions").Element("topAdExtension").Element("adDomainClicks").Value ?? "",
        private static String TryGetTopAdClicks(XElement element)
        {
            String TopAdClicks;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("topAdExtension") != null
                && element.Element("Extensions").Element("topAdExtension").Element("adDomainClicks") != null)
            {
                TopAdClicks = element.Element("Extensions").Element("topAdExtension").Element("adDomainClicks").Value;
            }
            else
            {
                TopAdClicks = "";
            }
            return TopAdClicks;
        }

        // 5. Security Badge
        // SecurityProvider = x.Element("Extensions").Element("securityBadgeAnnotation").Element("provider").Value ?? "",
        private static String TryGetSecurityProvider(XElement element)
        {
            String SecurityProvider;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("securityBadgeAnnotation") != null
                && element.Element("Extensions").Element("securityBadgeAnnotation").Element("provider") != null)
            {
                SecurityProvider = element.Element("Extensions").Element("securityBadgeAnnotation").Element("provider").Value;
            }
            else
            {
                SecurityProvider = "";
            }
            return SecurityProvider;
        }

        // 6. Brand Annotations
        // AdOfficialBrand = x.Element("Extensions").Element("brandAnnotation").Element("brandTitle").Value ?? "",
        private static String TryGetAdOfficialBrand(XElement element)
        {
            String AdOfficialBrand;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("brandAnnotation") != null
                && element.Element("Extensions").Element("brandAnnotation").Element("brandTitle") != null)
            {
                AdOfficialBrand = element.Element("Extensions").Element("brandAnnotation").Element("brandTitle").Value;
            }
            else
            {
                AdOfficialBrand = "";
            }
            return AdOfficialBrand;
        }

        // 7. Smart Annotations
        // SmartAdAnnotation = x.Element("Extensions").Element("SmartAnnotation").Element("Phrases").Element("Phrase").Value ?? ""
        private static String TryGetSmartAdAnnotation(XElement element)
        {
            String SmartAdAnnotation;
            if (element.Element("Extensions") != null
                && element.Element("Extensions").Element("SmartAnnotation") != null
                && element.Element("Extensions").Element("SmartAnnotation").Element("Phrases") != null
                && element.Element("Extensions").Element("SmartAnnotation").Element("Phrases").Element("phrase") != null)
            {
                SmartAdAnnotation = element.Element("Extensions").Element("SmartAnnotation").Element("Phrases").Element("Phrase").Value;
            }
            else
            {
                SmartAdAnnotation = "";
            }
            return SmartAdAnnotation;
        }
    }
}
