﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Web.Mvc;
using Amazon;
using Amazon.Runtime;
using Chronos.AWS;
using Chronos.Configuration;
using Mobile.Search.Providers;
using Mobile.Search.Shared;
using Mobile.Search.Shared.Extensions;
using Mobile.Search.Shared.Models;
using Mobile.Search.Web.Extensions;
using Mobile.Search.Web.Helpers;
using Mobile.Search.Web.Models;
using Newtonsoft.Json.Linq;

namespace Mobile.Search.Web.Controllers
{
    public class ResultsController : Controller
    {
        private static readonly IUrlHelper UrlHelper = new SearchUrlHelper();

        private static readonly string AWSAccessKey = ConfigUtilities.GetAppSetting("AWSAccessKey");
        private static readonly string AWSSecretKey = ConfigUtilities.GetAppSetting("AWSSecretKey");

        // GET: /results/{system}/json
        public ActionResult Json(JsonRequest resultsRequest)
        {
            var errorResponse = VerifyRequest(resultsRequest);
            if (errorResponse != null)
            {
                return Json(errorResponse, JsonRequestBehavior.AllowGet);
            }

            var sourceTag = GetSourceTag(resultsRequest.System, resultsRequest.SourceTagId);
            var typeTag = MakeD2STypeTag(resultsRequest.QuizId, resultsRequest.PubId, resultsRequest.SubPub,
                resultsRequest.Keyword, resultsRequest.TypeTagSuffix);
            var searchRequest = new SearchRequest
            {
                BiddedResults = 5,
                Ip = resultsRequest.Ip ?? RequestHelper.GetIp(HttpContext),
                NonBiddedResults = 0,
                Query = resultsRequest.Keyword,
                ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                SourceTag = sourceTag,
                Timestamp = DateTime.UtcNow,
                UseFakeUserAgent = false,
                TypeTag = typeTag,
                UserAgent = resultsRequest.UserAgent ?? Request.UserAgent,
            };
            var searchResults = YahooProvider.GetListings(searchRequest);
            var requestDict = Request.QueryString.ToDictionary();

            var debug = requestDict["debug"];
            bool isDebug = false;
            if (!string.IsNullOrEmpty(debug) && debug == "true") isDebug = true;
            Tracker.TrackSearch(searchRequest, searchResults, requestDict, isDebug);

            searchResults.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
              resultsRequest.ClickWrapper, requestDict, listing, searchResults);
            var templateVars = new TemplateVariables
            {
                Footer = "",
                Hotspot = "",
                Keyword = resultsRequest.Keyword,
            };

            PopulateTemplateVarsWithListings(searchResults.Listings.ToList(),
                searchResults.ClickUrlFunc, ref templateVars);

            var dynamoClient = new Dynamo(new BasicAWSCredentials(AWSAccessKey, AWSSecretKey), RegionEndpoint.USWest2);
            var json = dynamoClient.Get("prd_2click_templates", "templateKey", resultsRequest.TemplateKey);
            if (string.IsNullOrEmpty(json))
            {
                return Json(new
                {
                    Status = "Error",
                    Message = "Couldn't find a template with the key: " + resultsRequest.TemplateKey
                }, JsonRequestBehavior.AllowGet);
            }

            var jsonObject = JObject.Parse(json);
            var html = (string)jsonObject["html"];
            if (html != null)
            {
                var renderedTemplate = Nustache.Core.Render.StringToString(html, templateVars);
                return Json(new
                {
                    Status = "Success",
                    Html = renderedTemplate,
                    ResultsCount = searchResults.Listings != null && searchResults.Listings.Any()
                            ? searchResults.Listings.Count()
                            : 0,
                    SourceTag = searchResults.SourceTag,
                    TypeTag = searchResults.TypeTag
                }, JsonRequestBehavior.AllowGet);
            }

            return Json(new
            {
                Status = "Error",
                Message = "No html found within the template: " + resultsRequest.TemplateKey
            }, JsonRequestBehavior.AllowGet);

        }

        private string MakeD2STypeTag(int quizId, int pubId, string subPub, string keyword, string typeTagSuffix)
        {
            return string.Format("qu:{0}_pub:{1}_subpub:{2}_kw:{3}", quizId, pubId, subPub, keyword) + (typeTagSuffix ?? "");
        }

        private void PopulateTemplateVarsWithListings(List<Listing> listings, Func<Listing,string> clickUrlFunc, ref TemplateVariables templateVars)
        {
            if (listings.Count > 0)
            {
                templateVars.Listing1Title = listings[0].Title;
                templateVars.Listing1Advertiser = listings[0].SiteHost;
                templateVars.Listing1Description = listings[0].Description;
                templateVars.Listing1ClickUrl = clickUrlFunc(listings[0]);
            }
            if (listings.Count > 1)
            {
                templateVars.Listing2Title = listings[1].Title;
                templateVars.Listing2Advertiser = listings[1].SiteHost;
                templateVars.Listing2Description = listings[1].Description;
                templateVars.Listing2ClickUrl = clickUrlFunc(listings[1]);
            }
            if (listings.Count > 2)
            {
                templateVars.Listing3Title = listings[2].Title;
                templateVars.Listing3Advertiser = listings[2].SiteHost;
                templateVars.Listing3Description = listings[2].Description;
                templateVars.Listing3ClickUrl = clickUrlFunc(listings[2]);
            }
            if (listings.Count > 3)
            {
                templateVars.Listing4Title = listings[3].Title;
                templateVars.Listing4Advertiser = listings[3].SiteHost;
                templateVars.Listing4Description = listings[3].Description;
                templateVars.Listing4ClickUrl = clickUrlFunc(listings[3]);
            }
            if (listings.Count > 4)
            {
                templateVars.Listing5Title = listings[4].Title;
                templateVars.Listing5Advertiser = listings[4].SiteHost;
                templateVars.Listing5Description = listings[4].Description;
                templateVars.Listing5ClickUrl = clickUrlFunc(listings[4]);
            }
        }

        private string GetSourceTag(string system, int sourceTagid)
        {
            var st = SourceTagMappingHelper.Instance.GetSourceTagMappings(sourceTagid, system);
            if (st == null || st.Status == 0)
            {
                return "adknowledge_search";
            }
            return st.SourceTag;
        }

        private object VerifyRequest(JsonRequest resultsRequest)
        {
            bool valid = true;
            List<string> errorMessages = new List<string>();
             if (resultsRequest.Keyword == default(string))
             {
                 valid = false;
                 errorMessages.Add("Keyword parameter not specified");
            }
            if (resultsRequest.PubId == default(int))
            {
                valid = false;
                errorMessages.Add("PubId parameter not specifiied");
            }
            if (resultsRequest.QuizId == default(int))
            {
                valid = false;
                errorMessages.Add("QuizId parameter not specifiied");
            }
            if (resultsRequest.SourceTagId == default(int))
            {
                valid = false;
                errorMessages.Add("SourceTagId parameter not specifiied");
            }
            if (resultsRequest.TemplateKey == default(string))
            {
                 valid = false;
                errorMessages.Add("TemplateKey parameter not specifiied");
            }
            if (resultsRequest.SubPub == default(string))
            {
                valid = false;
                errorMessages.Add("SubPub parameter not specified");
            }

            if (!valid)
            {
                return new
                {
                    Status = "Error",
                    Message = string.Join("\n", errorMessages)
                };
            }
            return null;
        }
    }


    /// <summary>
    /// results/{system}/{outputFormat}
    /// http://mobsearch.adk-mobile/results/d2s/json 
    /// </summary>
    public class JsonRequest
    {
        public string System { get; set; }

        [Required]
        public string Keyword { get; set; }

        [Required]
        public int PubId { get; set; }

        [Required]
        public string SubPub { get; set; }

        [Required]
        public int SourceTagId { get; set; }

        [Required]
        public int QuizId { get; set; }

        [Required]
        public string TemplateKey { get; set; }

        public string ClickWrapper { get; set; }
        public string TypeTagSuffix { get; set; }
        public string Ip { get; set; }
        public string UserAgent { get; set; }
    }
}