﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;
using System.Web.Hosting;
using System.Web.Mvc;
using Amazon;
using Amazon.Runtime;
using Chronos;
using Chronos.AWS;
using Chronos.Configuration;
using Mobile.Search.Providers;
using Mobile.Search.Shared;
using Mobile.Search.Shared.Helpers;
using Mobile.Search.Shared.Extensions;
using Mobile.Search.Shared.Models;
using Mobile.Search.Web.Extensions;
using Mobile.Search.Web.Helpers;
using Mobile.Search.Web.Models;
using Newtonsoft.Json.Linq;
using Serilog;
using ServiceStack;
using StatsdClient;

namespace Mobile.Search.Web.Controllers
{
    public class SearchController : Controller
    {
        private static readonly IUrlHelper UrlHelper = new SearchUrlHelper();

        private static readonly string AWSAccessKey = ConfigUtilities.GetAppSetting("AWSAccessKey");
        private static readonly string AWSSecretKey = ConfigUtilities.GetAppSetting("AWSSecretKey");

        /// <summary>
        /// Our last stop before sending user to destination.
        /// Form submit redirect page for cloaking.  Uses a form submit so if there's no javascript
        /// enabled the user can still click their way through (not sure if this is valid concern or not) 
        /// </summary>
        public ActionResult Index(FormSubmitModel model)
        {
            if (!string.IsNullOrEmpty(model.RUrl))
            {
                // means we already handled tracking and referrer stuff
                // just redirect now
                return Redirect(model.RUrl);
            }
            if (string.IsNullOrEmpty(model.U))
            {
                // invalid click
                return RedirectToAction("Index", "Home");
            }


            var dictionary = Request.QueryString.ToDictionary();
            dictionary.AddOrUpdate("ip",RequestHelper.GetIp(HttpContext));
            dictionary.AddOrUpdate("userAgent", Request.UserAgent);
            dictionary.AddOrUpdate("timestamp", DateTime.UtcNow.ToString("O"));

            HostingEnvironment.QueueBackgroundWorkItem(ct => RelayMessage(dictionary));
            HostingEnvironment.QueueBackgroundWorkItem(ct =>
            {
                Metrics.Time(() => Tracker.TrackClick(dictionary), "trackClick");
            });
            

            if (Request.UrlReferrer != null && Request.UrlReferrer.Host.Contains("goodfind.guru"))
            {
                return Redirect(model.U);
            }

            // this form auto fills the Url property and submits
            return View("FormSubmit",model);
        }

        private string ToQueryString(NameValueCollection nvc)
        {
            var array = (from key in nvc.AllKeys
                         from value in nvc.GetValues(key)
                         select string.Format("{0}={1}", HttpUtility.UrlEncode(key), HttpUtility.UrlEncode(value)))
                .ToArray();
            return "?" + string.Join("&", array);
        }

        public void RelayMessage(Dictionary<string,string> dictionary )
        {
            var url = "";
            try
                {
                    var nvc = new NameValueCollection();
                    foreach (var kvp in dictionary)
                    {
                        if (kvp.Key != "u")
                        { nvc.Add(kvp.Key.ToString(), kvp.Value.ToString());}
                    }
                    url = "http://track.minbox.email/n/ce" + ToQueryString(nvc);
                    var client = new WebClient();
                    var result = client.DownloadString(url);
                }
                catch (Exception x)
                {
                Log.Error("Error: " + x);
                Log.Error("RelayMessage URL: " + url);
            }
        }

        public ActionResult R(string u)
        {
            var dictionary = Request.QueryString.ToDictionary();
            dictionary.AddOrUpdate("ip", RequestHelper.GetIp(HttpContext));
            dictionary.AddOrUpdate("userAgent", Request.UserAgent);
            dictionary.AddOrUpdate("timestamp", DateTime.UtcNow.ToString("O"));

            Metrics.Time(() => Tracker.TrackClick(dictionary), "trackClick");

            return Redirect(u);
        }

        public ActionResult Reveal_Click()
        {
            return new HttpStatusCodeResult(200);
        }

        public ActionResult MultiTerms(string activeKeyword, string subid, string clickWrapper = null, string footer = null)
        {
            var footerHtml = "";
            if (!footer.IsNullOrEmpty())
            {
                try
                {
                    footerHtml = Encoding.UTF8.GetString(Convert.FromBase64String(footer));
                }
                catch (Exception x)
                {
                    Log.Error("MultiTerms: Error with base 64 decoding", x);
                }
            }

            var viewModel = new MultiTermsViewModel
            {
                ActiveKeyword = activeKeyword ?? "interior+design+school",
                Subid = subid,
                ClickWrapper = clickWrapper,
                FooterHtml = footerHtml
            };

            return View(viewModel);
        }
        public ActionResult Adstation(string keyword, string keywordId, int? adCount, string listId)
        {
            var model = new AdStation()
            {
                Keyword = keyword,
                KeywordId = keywordId,
                ListId = listId ?? "23273",
                AdCount = adCount ?? 3,
                Ip = RequestHelper.GetIp(HttpContext)

            };

            return View(model);
        }


        /// <summary>
        /// Partial View that shows a listing, used in templates where multi keywords are used
        /// </summary>
        public ActionResult Listing(string keyword,string subid, string layout = "purple", int type = 2, string clickWrapper = null, bool debug = false)
        {
            keyword = YahooTags.CleanTypeTag(keyword);
            if (string.IsNullOrEmpty(keyword))
            {
                Log.Error("Null Keyword passed");
                return new HttpStatusCodeResult(400, "The keyword query parameter cannot be null");
            }

            var requestDict = Request.QueryString.ToDictionary();
            requestDict.AddOrUpdate("subid", subid);
            var searchRequest = new SearchRequest
            {
                BiddedResults = 5,
                NonBiddedResults = 0,
                Ip = RequestHelper.GetIp(HttpContext),
                Query = HttpUtility.UrlDecode(keyword),
                ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                SourceTagId = type,
                TypeTag = keyword,
                UserAgent = Request.UserAgent,
                Timestamp = DateTime.UtcNow,
                // Always fake user agent for now
                UseFakeUserAgent = true
            };
            var searchResult = Providers.YahooProvider.GetListings(searchRequest);

            searchResult.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                clickWrapper ?? string.Empty, requestDict, listing, searchResult);

            Tracker.TrackSearch(searchRequest, searchResult, requestDict, debug);

            switch (layout.ToLower())
            {
                case "purple":
                default:
                    return PartialView("Purple", searchResult);
            } 
        }


        // GET : /Search/Frame?keyword=query&layout=myLayout
        /// <summary>
        /// The search page
        /// </summary>
        /// <param name="keyword">
        /// Required and used for the type tag and the search query
        /// UrlDecode(keyword) = search query,
        /// UrlDecode(keyword).Replace(' ', '+') = type tag
        /// </param>
        /// <param name="clickWrapper">
        /// For custom click tracking
        /// </param>
        /// <param name="c">The count of sponsored listings to attempt to retreive</param>
        /// <param name="layout">the razor layout to use (same thing as a template)</param>
        /// <param name="type">The type of search (this determines the source tag)</param>
        public ActionResult Frame(string keyword,string clickWrapper = "",int c = 5,string layout = "google",
            int type = 1, string footer = null, bool debug = false, string sourceTag = null,
            string typeTagSuffix = null)
        {
            typeTagSuffix = typeTagSuffix ?? string.Empty;

            if (!footer.IsNullOrEmpty())
            {
                try
                {
                    ViewBag.FooterHtml = Encoding.UTF8.GetString(Convert.FromBase64String(footer));
                }
                catch (Exception x)
                {
                    Log.Error("Frame: " + x);
                }
            }
            keyword = YahooTags.CleanTypeTag(keyword);
            if (string.IsNullOrEmpty(keyword))
            {
                Log.Error("Null Keyword passed");
                return new HttpStatusCodeResult(400, "The keyword query parameter cannot be null");
            }

            var requestDict = Request.QueryString.ToDictionary();
            var searchRequest = new SearchRequest
            {
                BiddedResults = c,
                NonBiddedResults = 0,
                Ip = RequestHelper.GetIp(HttpContext),
                Query = HttpUtility.UrlDecode(keyword),
                ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                SourceTagId = type,
                TypeTag = keyword + typeTagSuffix,
                UserAgent = Request.UserAgent,
                SourceTag = sourceTag,
                Timestamp = DateTime.UtcNow,
                // Always fake user agent for now
                UseFakeUserAgent = true
            };
            var searchResult = Providers.YahooProvider.GetListings(searchRequest);

            searchResult.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                clickWrapper, requestDict, listing, searchResult);

            Tracker.TrackSearch(searchRequest, searchResult, requestDict, debug);

            switch (layout.ToLower())
            {
                case "redblack":
                    return View("RedBlack", searchResult);
                case "google_single":
                    return View("GoogleSingle", searchResult);
                case "multi_listing":
                    return View("MultiListing", searchResult);
                case "purple":
                    return View("Purple", searchResult);
                case "google":
                default:
                    return View("Google", searchResult);
            }
        }

        public ActionResult Html(string templateKey)
        {
            Response.AppendHeader("Access-Control-Allow-Origin", "*");
            var dynamoClient = new Dynamo(new BasicAWSCredentials(AWSAccessKey, AWSSecretKey), RegionEndpoint.USWest2);
            var json = dynamoClient.Get("prd_2click_templates", "templateKey",templateKey);
            if (string.IsNullOrEmpty(json))
            {
                return HttpNotFound("Couldn't find a template with the key: " + templateKey);
            }
            var jsonObject = JObject.Parse(json);
            var html = (string) jsonObject["html"];
            if (html != null)
            {
                return Content(html);
            }
            return new HttpStatusCodeResult(400, "Couldn't find template html for key: " + templateKey);
        }

        public class TemplateReturnObj
        {
            public string Html { get; set; }
            public string Status { get; set; }
            public int? ResultsCount { get; set; }
        }

        // this is ugly, but need this fixed quick, just pass the new tag in
        private string ToQueryString(NameValueCollection nvc, string newPartnerTag, string newTypeTag)
        {
            var list = (from key in nvc.AllKeys
                from value in nvc.GetValues(key)
                select string.Format("{0}={1}", HttpUtility.UrlEncode(key), HttpUtility.UrlEncode(value)))
                .ToList();

            for (var i=0; i < list.Count; i++)
            {
                if (list[i].StartsWith("partner="))
                    list[i] = "partner=" + HttpUtility.UrlEncode(newPartnerTag);

                if (list[i].StartsWith("typeTag="))
                    list[i] = "typeTag=" + HttpUtility.UrlEncode(newTypeTag);
            }


            return "?" + string.Join("&", list);
        }

        /// <summary>
        /// For D2S only right now
        /// </summary>
        /// <param name="templateKey"></param>
        /// <param name="keyword"></param>
        /// <param name="clickWrapper">custom tracking, if not included we use the d2s one</param>
        /// <param name="footer">the footer has to be base64 encoded</param>
        /// <param name="c">number of results to retrieve</param>
        /// <param name="debug"></param>
        /// <param name="subject"></param>
        /// <param name="hotspot"></param>
        /// <param name="partner">this is the source tag</param>
        /// <param name="typeTag"></param>
        public void TemplateJson(string templateKey, string keyword = null, string clickWrapper = null,
            string footer = null, int c = 5, bool debug = false, string subject = null, string hotspot = null,
            string partner = null, string typeTag = null, string userAgent = null, string ip = null, string quiz_id = null, string question_id = null)
        {

            // delete this method forever as soon as we get the new quiz thing working
            // this shit is a mess don't bother trying cleaning it

            var returnObj = new TemplateReturnObj();
            Response.ContentType = "application/json";
            //typeTagSuffix = typeTagSuffix ?? string.Empty;
            Response.AppendHeader("Access-Control-Allow-Origin", "*");
            var templateVars = new TemplateVariables
            {
                Footer = footer,
                Hotspot = hotspot,
                Subject = subject
            };
            if (keyword != null)
            {
                // I don't remember why we're decoding this...it should come in decoded
                templateVars.Keyword = WebUtility.UrlDecode(keyword);
            }

            try
            {
                // this is iAd, try to change the type tag to new format
                // testing
                if (partner == "adknowledge_d2s_ilotto_search" && !string.IsNullOrEmpty(Request["quiz_id"]) &&
                    !string.IsNullOrEmpty(Request["publisherid"])
                    && (!string.IsNullOrEmpty(Request["list_id"]) || !string.IsNullOrEmpty(Request["subpub"])) &&
                    !string.IsNullOrEmpty(Request["keyword"]))
                {
                    var subpub = string.IsNullOrEmpty(Request["subPub"]) ? Request["list_id"] : Request["subpub"];
                    typeTag = "qu:" + Request.Form["quiz_id"] + "_pub:" + Request["publisherid"] + "_subpub:" +
                              subpub + "_kw:" + keyword;

                }
            }
            catch (Exception x)
            {
                Log.Error("TemplateJson: Error trying to create type tag for iad test", x);
            }

            // pretty sure this can't throw, but i'm paranoid
            // should only choose a random partner if the partner 
            // tag equals adknowledge_d2s_nesn_search (taboola mobile)
            try
            {
                partner = RandomizeParther(partner);
                Metrics.Counter("searches.sourcetag." + partner);
            }
            catch (Exception x)
            {
                Log.Error("TemplateJson: IMPOSSIBLEEEE!", x);
            }

            if (clickWrapper == null)
            {
                var url = "http://paidclicks.imp.d2s.adk-mobile.com/" + ToQueryString(Request.Form, partner, typeTag);
                url = url.AddQueryParam("ip", RequestHelper.GetIp(HttpContext));
                url = url.AddQueryParam("url", "{URL}", false);
                url = url.AddQueryParam("event", "paidclicks");
                clickWrapper = url;
            }

            if (keyword != null)
            {
                keyword = YahooTags.CleanTypeTag(keyword);
                if (string.IsNullOrEmpty(keyword))
                {
                    Log.Error("Null Keyword passed");

                    returnObj.Status = "Error";
                    returnObj.Html = "<p>The keyword query parameter cannot be null</p>";

                    Response.Write(returnObj.ToJson());
                    Response.End();
                    return;
                }
            }


           
            var requestDict = Request.QueryString.ToDictionary();
            var searchRequest = new SearchRequest
            {
                BiddedResults = c,
                NonBiddedResults = 0,
                Ip = ip ?? RequestHelper.GetIp(HttpContext),
                Query = WebUtility.UrlDecode(keyword),
                ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                SourceTagId = 1,
                SourceTag = string.IsNullOrEmpty(partner) ? "adknowledge_search" : partner,
                TypeTag = typeTag,
                UserAgent = userAgent ?? Request.UserAgent,
                Timestamp = DateTime.UtcNow,
                UseFakeUserAgent = false,
            };


            var searchResult = Metrics.Time(() => Providers.YahooProvider.GetListings(searchRequest), "yahooResultsTime");

             searchResult.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                clickWrapper, requestDict, listing, searchResult);

            Func<string,string> keywordToNextUrl = (varStr) =>
            {
                var tk = default(string);
                var kw = default(string);
                var vars = varStr.Split(new[] {":"}, StringSplitOptions.RemoveEmptyEntries);
                if (vars.Length > 0)
                    tk = vars[0];
                if (vars.Length > 1)
                    kw = vars[1];
                return UrlHelper.GetAuthority(Request) + Url.Action("Template",
                    new
                    {
                        templateKey = tk,
                        keyword = kw ?? keyword,
                        clickWrapper,
                        footer,
                        c,
                        debug,
                        subject,
                        hotspot
                    });
            };

            Func<string, string, string> wrapNextUrl = (cWrapper, nextUrl) =>
            {
                if (string.IsNullOrEmpty(cWrapper)) return nextUrl;
                var endUrl = Regex.Replace(cWrapper, "{URL}", WebUtility.UrlEncode(nextUrl), RegexOptions.IgnoreCase);
                endUrl.AddQueryParam("searchId", searchResult.SearchId.ToString("N"));
                return endUrl;
            };

            templateVars.WrapUrl = s => Tracker.WrapPlainUrl(clickWrapper, s, searchResult.SearchId);
            templateVars.TemplateUrl = vars => wrapNextUrl(clickWrapper, keywordToNextUrl(vars));

            Metrics.Time(() => Tracker.TrackSearch(searchRequest, searchResult, requestDict, debug), "trackSearch");

            var listings = searchResult.Listings.ToList();
            if (listings.Count > 0)
            {
                templateVars.Listing1Title = listings[0].Title;
                templateVars.Listing1Advertiser = listings[0].SiteHost;
                templateVars.Listing1Description = listings[0].Description;
                templateVars.Listing1AdvertiserUrl = listings[0].GetAdvertiserUrl();
                templateVars.Listing1ListingUrl = listings[0].ClickUrl;
                templateVars.Listing1ClickUrl = searchResult.ClickUrlFunc(listings[0]);
            }
            if (listings.Count > 1)
            {
                templateVars.Listing2Title = listings[1].Title;
                templateVars.Listing2Advertiser = listings[1].SiteHost;
                templateVars.Listing2Description = listings[1].Description;
                templateVars.Listing2AdvertiserUrl = listings[1].GetAdvertiserUrl();
                templateVars.Listing2ListingUrl = listings[1].ClickUrl;
                templateVars.Listing2ClickUrl = searchResult.ClickUrlFunc(listings[1]);
            }
            if (listings.Count > 2)
            {
                templateVars.Listing3Title = listings[2].Title;
                templateVars.Listing3Advertiser = listings[2].SiteHost;
                templateVars.Listing3Description = listings[2].Description;
                templateVars.Listing3AdvertiserUrl = listings[2].GetAdvertiserUrl();
                templateVars.Listing3ListingUrl = listings[2].ClickUrl;
                templateVars.Listing3ClickUrl = searchResult.ClickUrlFunc(listings[2]);
            }
            if (listings.Count > 3)
            {
                templateVars.Listing4Title = listings[3].Title;
                templateVars.Listing4Advertiser = listings[3].SiteHost;
                templateVars.Listing4Description = listings[3].Description;
                templateVars.Listing4AdvertiserUrl = listings[3].GetAdvertiserUrl();
                templateVars.Listing4ListingUrl = listings[3].ClickUrl;
                templateVars.Listing4ClickUrl = searchResult.ClickUrlFunc(listings[3]);
            }
            if (listings.Count > 4)
            {
                templateVars.Listing5Title = listings[4].Title;
                templateVars.Listing5Advertiser = listings[4].SiteHost;
                templateVars.Listing5Description = listings[4].Description;
                templateVars.Listing5AdvertiserUrl = listings[4].GetAdvertiserUrl();
                templateVars.Listing5ListingUrl = listings[4].ClickUrl;
                templateVars.Listing5ClickUrl = searchResult.ClickUrlFunc(listings[4]);
            }
            var dynamoClient = new Dynamo(new BasicAWSCredentials(AWSAccessKey, AWSSecretKey), RegionEndpoint.USWest2);
            var json = dynamoClient.Get("prd_2click_templates", "templateKey",templateKey);
            if (string.IsNullOrEmpty(json))
            {
                Log.Error("Couldn't find a template with key: @templateKey", templateKey);
                returnObj.Status = "Error";
                returnObj.Html = "<p>Couldn't find a template with the key: " + templateKey + "</p>";
                Response.ClearContent();
                Response.Write(returnObj.ToJson());
                Response.End();
                return;
            }
            var jsonObject = JObject.Parse(json);
            var html = (string) jsonObject["html"];
            if (html != null)
            {
                var renderedTemplate = Nustache.Core.Render.StringToString(html, templateVars);
                returnObj.Status = "Success";
                returnObj.Html = renderedTemplate;
                returnObj.ResultsCount = searchResult.Listings.Count();
                Response.ClearContent();
                Response.Write(returnObj.ToJson());
                Response.End();
            }
            returnObj.Status = "Error";
            returnObj.Html = "<p>Couldn't find template html for key: " + templateKey + "</p>";
            Response.ClearContent();
            Response.Write(returnObj.ToJson());
            Response.End();
        }

        /// <summary>
        /// This only should happen for taboola mobile
        /// </summary>
        /// <param name="partner"></param>
        /// <returns></returns>
        private string RandomizeParther(string partner)
        {
            if (partner == "adknowledge_d2s_nesn_search")
            {
                return "adknowledge_d2s_quizrocket_search";
            }
            return partner;
        }

        public ActionResult TemplateKwb(string templateKey, string uuid, string clickWrapper, string footer = null,
            string subject = null, string hotspot = null, string typeTagSuffix = null, string sourceTagOverride = null)
        {
            sourceTagOverride = string.IsNullOrEmpty(sourceTagOverride)
                ? "adknowledge_mobile_2click_newsletter_search"
                : sourceTagOverride;

            typeTagSuffix = typeTagSuffix ?? string.Empty;
            Response.AppendHeader("Access-Control-Allow-Origin", "*");

            var templateVars = new KeywordBlockVariables
            {
                Footer = footer,
                Hotspot = hotspot,
                Subject = subject,
                Uuid = Request["uuid"],
                EmailId = Request["email_id"] ?? Request["emailId"],
                UserClass = Request["user_class"] ?? Request["user_class"],
                OfferId = Request["offer_id"] ?? Request["offerId"],
                Product = Request["product"],
                TemplateName = templateKey
            };

            var keywords = KeywordBlockHelper.GetKeywordsForUser(uuid);


            clickWrapper = IncrementClickCountOnClickWrapper(clickWrapper, baseValue: "kwb_click_");
            clickWrapper = AddKeywordMacroToClickWrapper(clickWrapper);
            var html = string.Empty;
            if (!TemplateCache.Instance.TryGetTemplateHtml(templateKey, out html))
            {
                return HttpNotFound(string.Format("Couldn't retreive template with key {0}", templateKey));
            }

            var requestDict = Request.QueryString.ToDictionary();
            // replace the existing one with this to fix the clicks
            requestDict["sourceTagOverride"] = sourceTagOverride;
            requestDict.Remove("footer");

            var ip = Request["ip"] ?? RequestHelper.GetIp(HttpContext);
            Func<string,string> getResultUrl = keyword => UrlHelper.GetAuthority(Request) + Url.Action("Template",
                    new
                    {
                        templateKey = "1-click search intent",
                        keyword,
                        clickWrapper,
                        footer,
                        c = 2,
                        subject,
                        hotspot
                    });


            if (keywords.Any())
            {
                templateVars.Keyword1 = new KeywordBlockKeyword
                {
                    Keyword = keywords[0],
                    Listing1ClickUrl = Tracker.WrapPlainUrl(clickWrapper, getResultUrl(keywords[0]), null, new Dictionary<string, string> { { "keyword", keywords[0]}})
                };
            }
            if (keywords.Count > 1)
            {
                templateVars.Keyword2 = new KeywordBlockKeyword
                {
                    Keyword = keywords[1],
                    Listing1ClickUrl = Tracker.WrapPlainUrl(clickWrapper, getResultUrl(keywords[1]), null, new Dictionary<string, string> { { "keyword", keywords[1]}})
                };
            }
            if (keywords.Count > 2)
            {
                templateVars.Keyword3 = new KeywordBlockKeyword
                {
                    Keyword = keywords[2],
                    Listing1ClickUrl = Tracker.WrapPlainUrl(clickWrapper,getResultUrl(keywords[2]), null, new Dictionary<string, string> { { "keyword", keywords[2]}})
                };
            }

            var renderedTemplate = Nustache.Core.Render.StringToString(html, templateVars);
            return Content(renderedTemplate);
        }

        private string AddKeywordMacroToClickWrapper(string clickWrapper)
        {
            if (clickWrapper == null)
            {
                return null;
            }
            const string pattern = @"keyword=([^&]*)";
            var matchSuccess = false;
            clickWrapper = Regex.Replace(clickWrapper, pattern, m =>
            {
                matchSuccess = true;

                return "keyword={KEYWORD}";
            });
            if (!matchSuccess)
            {
                clickWrapper = clickWrapper.AddQueryParam("keyword", "{KEYWORD}", false);
            }
            return clickWrapper;
        }

        public ActionResult TemplateRelated(string templateKey, string uuid, string clickWrapper = null, string footer = null,
            string subject = null, string hotspot = null, string typeTagSuffix = null, string sourceTagOverride = null)
        {
            // kill this for now
            return Content("");

            // sometimes this is passed as an empty string
            /*sourceTagOverride = string.IsNullOrEmpty(sourceTagOverride) ? "adknowledge_mobile_2click_newsletter_search" : sourceTagOverride;

            typeTagSuffix = typeTagSuffix ?? string.Empty;
            Response.AppendHeader("Access-Control-Allow-Origin", "*");
            var templateVars = new KeywordBlockVariables 
            {
                Footer = footer,
                Hotspot = hotspot,
                Subject = subject,
                Uuid = Request["uuid"],
                EmailId = Request["email_id"] ?? Request["emailId"],
                UserClass = Request["user_class"] ?? Request["user_class"],
                OfferId = Request["offer_id"] ?? Request["offerId"],
                Product = Request["product"],
                TemplateName = templateKey
            };

            var keywords = KeywordBlockHelper.GetKeywordsForUser(uuid);


            clickWrapper = IncrementClickCountOnClickWrapper(clickWrapper);
            var html = string.Empty;
            if (!TemplateCache.Instance.TryGetTemplateHtml(templateKey, out html))
            {
                return HttpNotFound(string.Format("Couldn't retreive template with key {0}", templateKey));
            }

            var requestDict = Request.QueryString.ToDictionary();
            // replace the existing one with this to fix the clicks
            requestDict["sourceTagOverride"] = sourceTagOverride;
            requestDict.Remove("footer");

            var ip = Request["ip"] ?? RequestHelper.GetIp(HttpContext);
            for (var index = 0; index < keywords.Count; index++)
            {
                var keyword = keywords[index];
                var searchRequest = new SearchRequest
                {
                    BiddedResults = 2,
                    NonBiddedResults = 0,
                    Ip = ip,
                    Query = WebUtility.UrlDecode(keyword),
                    ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                    SourceTag = string.IsNullOrEmpty(sourceTagOverride) ? "adknowledge_mobile_2click_newsletter_search" : sourceTagOverride,
                    TypeTag = keyword + typeTagSuffix,
                    UserAgent = Request.UserAgent,
                    Timestamp = DateTime.UtcNow,
                    UseFakeUserAgent = false,
                };

                var result = YahooProvider.GetListings(searchRequest);
                Tracker.TrackSearch(searchRequest, result, requestDict);
                result.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                   clickWrapper, requestDict, listing, result);
                var prop = templateVars.GetType().GetProperty("Keyword" + (index+1));
                var listings = result.Listings.ToArray();
                var kwbVar = new KeywordBlockKeyword
                {
                    Keyword = keywords[index].Replace("+", " "),
                    Listing1Title = listings[0].Title,
                    Listing1Advertiser = listings[0].SiteHost,
                    Listing1ClickUrl = result.ClickUrlFunc(listings[0]),
                    Listing1Description = listings[0].Description,
                    Listing2Title = listings.Length >= 2 ? listings[1].Title : string.Empty,
                    Listing2Advertiser = listings.Length >= 2 ? listings[1].SiteHost : string.Empty,
                    Listing2Description = listings.Length >= 2 ? listings[1].Description : string.Empty,
                    Listing2ClickUrl = listings.Length >= 2 ? result.ClickUrlFunc(listings[0]) : null,
                };

                prop.SetValue(templateVars, kwbVar);
            }

            templateVars.WrapUrl = s => Tracker.WrapPlainUrl(clickWrapper, Nustache.Core.Render.StringToString(s, templateVars), null);
            var renderedTemplate = Nustache.Core.Render.StringToString(html, templateVars);

            return Content(renderedTemplate);
            */
        }

        public ActionResult Template(string templateKey, string keyword = null, string clickWrapper = null,
            string footer = null, int c = 5, bool debug = false, string subject = null, string hotspot = null,
            string typeTagSuffix = null, string sourceTagOverride = null, int sourceTagIdOverride = 0,
            int nbr = 0, bool storeClickURLs = false, string typeTag = null, int? productId = null, string product = null)
        {
            typeTagSuffix = typeTagSuffix ?? string.Empty;
            
            //convert product to product Id if we can
            if (!productId.HasValue && !string.IsNullOrEmpty(product))
            {
                var p = 0;
                if(int.TryParse(product, out p))
                {
                    productId = p;
                }
            }

            clickWrapper = IncrementClickCountOnClickWrapper(clickWrapper);

            // offer id typically isn't passed in, so try to pull it from the click wrapper
            var offerId = Request["offer_id"] ?? Request["offerId"];
            if (!string.IsNullOrEmpty(clickWrapper) && string.IsNullOrEmpty(offerId))
            {
                try
                {
                    var uri = new Uri(clickWrapper);
                    var nvc = HttpUtility.ParseQueryString(uri.Query);
                    offerId = nvc["offer_id"] ?? string.Empty;
                }
                catch (Exception x)
                {
                    Log.Error(x, "Template: Error trying to parse click wrapper and extract offer id {url}", clickWrapper);
                }
            }
            var templateVars = new TemplateVariables
            {
                Footer = footer,
                Hotspot = hotspot,
                Subject = subject,
                Keyword = keyword,
                Uuid = Request["uuid"],
                EmailId = Request["email_id"] ?? Request["emailId"] ?? Request["emailid"],
                UserClass = Request["user_class"] ?? Request["user_class"],
                OfferId = offerId,
                Product = Request["product"],
                TemplateName = templateKey,
                DateString = DateTime.UtcNow.Date.ToString("yyyyMMdd"),
                Name = Request["name"]
            };


            keyword = YahooTags.CleanTypeTag(keyword);
            if (string.IsNullOrEmpty(keyword))
            {
                Log.Error("Null Keyword passed");
                return new HttpStatusCodeResult(400, "The keyword query parameter cannot be null");
            }


            var html = string.Empty;
            if (!TemplateCache.Instance.TryGetTemplateHtml(templateKey, out html))
            {
                return HttpNotFound($"Couldn't retreive template with key {templateKey}");
            }

            var searchRegex = new Regex(@"\{\{&?#?Listing");
            var searchMatch = searchRegex.Match(html);
            var shouldSearch = searchMatch.Success;

            var searchIndeedRegex = new Regex(@"\{\{&?#?IndeedListing");
            var searchIndeedMatch = searchIndeedRegex.Match(html);
            var shouldSearchIndeed = searchIndeedMatch.Success;

            // Determines if this search should set the Taboola API flag to true
            var searchTaboolaRegex = new Regex(@"\{\{&?#?TaboolaListing");
            var searchTaboolaMatch = searchTaboolaRegex.Match(html);
            var shouldSearchTaboola = searchTaboolaMatch.Success;

            var msnOnly = false;
            string[] msnOnlyTestKeywords = {
                "asian+dating+personals", "veterans+benefits", "crockpot+chicken+recipes", "dinner+restaurant+menus",
                "best+cell+phone+deals", "cell+phone+monitoring", "no+contract+cell+phones", "plus+size+clothing",
                "security+camera+system", "senior+dating+website", "new+iphone", };
            if (msnOnlyTestKeywords.Contains(keyword))
            {
                var rnd = new Random();
                msnOnly = rnd.Next(0, 100) < 50;
                if (msnOnly)
                {
                    typeTagSuffix += "_35";
                }
                else
                {
                    typeTagSuffix += "_32";
                }
            }

           //if we passed in a type tag, then we should use it.  Otherwise build our own
            if (string.IsNullOrEmpty(typeTag))
            {
                // how type tags work (old way)
                // _<suffix> is optional
                // all keys start with an underscore, then the key name, then a colon, then the value
                // <keyword>_<suffix>_<key>:<value>_<key2>:<value2>

                //new way for type tags:
                //key:value;

                typeTag = string.Format("k:{0};", keyword);
                if (!string.IsNullOrEmpty(templateVars.OfferId))
                {
                    typeTag += string.Format("oid:{0};", templateVars.OfferId);
                }
                if (!string.IsNullOrEmpty(templateVars.Product))
                {
                    //make sure this is a number
                    var pdid = 0;
                    if (int.TryParse(templateVars.Product, out pdid))
                    {
                        typeTag += string.Format("pdid:{0};", pdid);
                    }

                    
                }

                //TODO for now we'll ignore suffix (it was the template id from wasabi) but may want to add the ability for wasabi to add one later
                if (!string.IsNullOrEmpty(typeTagSuffix))
                {
                    typeTag += string.Format("sx:{0};", typeTagSuffix);
                }


            }
            var requestDict = Request.QueryString.ToDictionary();

            requestDict.Remove("footer");
            var sourceTag = sourceTagOverride;
            bool is_manual_offer_blast = false;
            if (sourceTagOverride == "adknowledge_mobile_2click_newsletter_search")
            {
                is_manual_offer_blast = true;
            }
            //TODO this is a hard code for now, we can make this data driven and put it back in the calling systems
            //adknowledge_mobile_2click_e_search = launcher email
            //xml_reachmobi_launcher = launcher kwb
            //adknowledge_mobile_2click_d_search = day0
            //adknowledge_mobile_2click_k_search = 2 click browser test
            if (productId.HasValue && (string.IsNullOrEmpty(sourceTagOverride) || 
                (!sourceTagOverride.Equals("adknowledge_mobile_2click_e_search",StringComparison.CurrentCultureIgnoreCase) 
                && !sourceTagOverride.ToLower().StartsWith("xml_reachmobi_launcher_") 
                && !sourceTagOverride.Equals("adknowledge_mobile_2click_d_search",StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("adknowledge_mobile_2click_k_search", StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("reachmobi_mobileinbox_Go", StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("reachmobi_mobileinbox_LearnMore", StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("reachmobi_mobileinbox_SeeIT", StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("reachmobi_mobileinbox_SeeNow", StringComparison.CurrentCultureIgnoreCase)
                && !sourceTagOverride.Equals("reachmobi_mobileinbox_ViewNow", StringComparison.CurrentCultureIgnoreCase)
                )) &&!is_manual_offer_blast )
            {
                //GET SOURCE TAG
                sourceTag = YahooSourceTagHelper.GetSourceTag(productId, sourceTagOverride);
            }


            var ip = Request["ip"] ?? RequestHelper.GetIp(HttpContext);
            var searchRequest = new SearchRequest
            {
                BiddedResults = c,
                NonBiddedResults = nbr,
                Ip = ip,
                Query = WebUtility.UrlDecode(keyword),
                ServeUrl = UrlHelper.GetAuthority(Request) + "/search",
                SourceTagId = sourceTagIdOverride,
                SourceTag = sourceTag,
                TypeTag = typeTag,
                UserAgent = Request.UserAgent,
                Timestamp = DateTime.UtcNow,
                UseFakeUserAgent = false,
            };

            SearchResult searchResult;
            if (shouldSearch)
            {
                // AK: look into fixing ClickUrl 
                var sw = System.Diagnostics.Stopwatch.StartNew();
                searchResult = Providers.YahooProvider.GetListings(searchRequest);
                RedisHelper.AddRequestTime("YahooSearch", sw.Elapsed.TotalMilliseconds);
                searchResult.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                    clickWrapper, requestDict, listing, searchResult);
            }
            else
            {
                searchResult = new SearchResult();
            }

            templateVars.Keyword = searchRequest.Query;

            Func<string,string> keywordToNextUrl = (varStr) =>
            {
                var tk = default(string);
                var kw = default(string);
                var vars = varStr.Split(new[] {":"}, StringSplitOptions.RemoveEmptyEntries);
                if (vars.Length > 0)
                    tk = vars[0];
                if (vars.Length > 1)
                    kw = vars[1];
                return UrlHelper.GetAuthority(Request) + Url.Action("Template",
                    new
                    {
                        templateKey = tk,
                        keyword = kw ?? keyword,
                        clickWrapper,
                        footer,
                        c,
                        debug,
                        subject,
                        hotspot
                    });
            };

            Func<string, string, string> wrapNextUrl = (cWrapper, nextUrl) =>
            {
                if (string.IsNullOrEmpty(cWrapper)) return nextUrl;
                var endUrl = Regex.Replace(cWrapper, "{URL}", WebUtility.UrlEncode(nextUrl), RegexOptions.IgnoreCase);
                if (searchResult.SearchId != null)
                {
                    endUrl.AddQueryParam("searchId", searchResult.SearchId.ToString("N"));
                }
                return endUrl;
            };

            templateVars.WrapUrl = s => Tracker.WrapPlainUrl(clickWrapper, s, searchResult.SearchId);
            templateVars.TemplateUrl = vars => wrapNextUrl(clickWrapper, keywordToNextUrl(vars));
            templateVars.Ip = searchRequest.Ip;

            if (shouldSearch)
            {
                try
                {
                    HostingEnvironment.QueueBackgroundWorkItem(ct => Tracker.TrackSearch(searchRequest, searchResult, requestDict, debug)); // 3 versions of the params... 
                }
                catch (Exception x)
                {
                    Log.Error("Template: Error tracking search", x);
                }

                //var listings = searchResult.Listings.ToList();
                var advertiserBlacklist = GetAdvertiserBlacklistForKeyword(keyword);
                var listings = searchResult.Listings.Where(x =>
                {
                    var advertiserUrl = x.GetAdvertiserUrl().ToLower();
                    var blacklisted = advertiserBlacklist.Any(b =>
                    {
                        var pattern = "^" + Regex.Escape(b).Replace(@"\*", ".*") + "$";
                        Regex regex = new Regex(pattern);
                        return regex.Match(advertiserUrl).Success;
                    });
                    if (blacklisted)
                    {
                        return false;
                    }
                    if (msnOnly && !x.ClickUrl.Contains("r.msn.com"))
                    {
                        return false;
                    }

                    if (String.IsNullOrEmpty(x.GetTitle()))
                    {
                        System.Diagnostics.Debug.WriteLine("TITLE IS" + x.GetTitle());
                        return false;
                    }
                    return true;
                }).ToList();
                
                var clickUrls = new Dictionary<string, string>();
                templateVars.Listings = new List<ListingTemplateVariables>();
                for (var i = 0; i < Math.Min(listings.Count, c + nbr); i++)
                {
                    var listing = new ListingTemplateVariables();
                    listing.ListingIndex = i + 1;
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex, listing);
                    templateVars.Listings.Add(listing);

                    listing.SiteLinks = new List<ListingTemplateVariables.SiteLink>();
                    for (var s = 0; s < 4; s++)
                    {
                        var siteLink = new ListingTemplateVariables.SiteLink();
                        siteLink.SiteLinkIndex = s + 1;
                        TrySetProperty(listing, "SiteLink" + siteLink.SiteLinkIndex, siteLink);
                        listing.SiteLinks.Add(siteLink);
                    }

                    if (i >= listings.Count)
                    {
                        continue;
                    }

                    listing.BiddedListing = listings[i].BiddedListing;
                    listing.Rank = listings[i].Rank;
                    listing.Title = listings[i].Title;
                    listing.Advertiser = listings[i].SiteHost;
                    listing.Description = listings[i].Description;
                    listing.AdvertiserUrl = listings[i].GetAdvertiserUrl();
                    listing.ListingUrl = listings[i].ClickUrl;
                    listing.ClickUrl = searchResult.ClickUrlFunc(listings[i]);
                    listing.FaviconUrl = listings[i].FaviconUrl;
                    listing.MerchRatingNumber = listings[i].MerchRatingNumber;
                    listing.MerchRatingUrl = listings[i].MerchRatingUrl;
                    listing.PhoneNumber = listings[i].PhoneNumber;
                    listing.LongTitle = listings[i].LongTitle;
                    listing.ImageUrl = listings[i].ImageUrl;

                    // === Yahoo! API Extensions Update 2017-10-11 === //

                    // 1. Callout
                    listing.CalloutAnnotation = listings[i].CalloutAnnotation;

                    // 2. Review
                    listing.ReviewSource = listings[i].ReviewSource;
                    listing.ReviewText = listings[i].ReviewText;

                    // 3. 4th Line
                    listing.Ad4thLine = listings[i].Ad4thLine;

                    // 4. Top Ad
                    listing.HasTopAdAnnotation = listings[i].HasTopAdAnnotation;
                    listing.TopAdDomain = listings[i].TopAdDomain;
                    listing.TopAdClicks = listings[i].TopAdClicks;

                    // 5. Security Badge
                    listing.SecurityProvider = listings[i].SecurityProvider;

                    // 6. Brand
                    listing.AdOfficialBrand = listings[i].AdOfficialBrand;

                    // 7. Smart Ad
                    listing.SmartAdAnnotation = listings[i].SmartAdAnnotation;

                    var listingTextLength = listing.Title.Length + listing.Description.Length;
                    if (listingTextLength > 100)
                    {
                        listing.DisplaySizePercentage = (int)((150 - Math.Min(listingTextLength, 150)) / 50.0 * 35.0) + 65;
                    }
                    else
                    {
                        listing.DisplaySizePercentage = 100;
                    }

                    for (var s = 0; s < Math.Min(listing.SiteLinks.Count, listings[i].SiteLinks.Count); s++)
                    {
                        var siteLink = listing.SiteLinks[s];
                        siteLink.Text = listings[i].SiteLinks[s].Text;
                        siteLink.Url = listings[i].SiteLinks[s].Url;
                        // hacky way to get around not passing click url through ClickUrlFunc
                        var listingClickUrl = listings[i].ClickUrl;
                        listings[i].ClickUrl = siteLink.Url;
                        siteLink.ClickUrl = searchResult.ClickUrlFunc(listings[i]);
                        listings[i].ClickUrl = listingClickUrl;

                        clickUrls["Listing" + listing.ListingIndex + "SiteLink" + siteLink.SiteLinkIndex + "ClickUrl"] = siteLink.ClickUrl;
                    }

                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "Title", listing.Title);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "Advertiser", listing.Advertiser);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "Description", listing.Description);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "AdvertiserUrl", listing.AdvertiserUrl);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "ListingUrl", listing.ListingUrl);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "ClickUrl", listing.ClickUrl);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "FaviconUrl", listing.FaviconUrl);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "MerchRatingNumber", listing.MerchRatingNumber);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "MerchRatingUrl", listing.MerchRatingUrl);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "PhoneNumber", listing.PhoneNumber);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "LongTitle", listing.LongTitle);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "ImageUrl", listing.ImageUrl);

                    // === Yahoo! API Extensions Update 2017-10-11 === //

                    // 1. Callout Extension
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "CalloutAnnotation", listing.CalloutAnnotation);

                    // 2. Review Extension
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "ReviewSource", listing.ReviewSource);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "ReviewText", listing.ReviewText);

                    // 3. 4th Line Annotation
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "Ad4thLine", listing.Ad4thLine);

                    // 4. Top Ad Annotation
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "HasTopAdAnnotation", listing.HasTopAdAnnotation);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "TopAdDomain", listing.TopAdDomain);
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "TopAdClicks", listing.TopAdClicks);

                    // 5. Security Badge
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "SecurityProvider", listing.SecurityProvider);

                    // 6. Brand Annotations
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "AdOfficialBrand", listing.AdOfficialBrand);

                    // 7. Smart Ad Annotation
                    TrySetProperty(templateVars, "Listing" + listing.ListingIndex + "SmartAdAnnotation", listing.SmartAdAnnotation);

                    if (string.IsNullOrEmpty(listing.ListingUrl) || !listing.ListingUrl.ToLower().StartsWith("http"))
                    {
                        Log.Warning("Null, empty, or invalid ListingUrl found for listing {listing}: of search query {query}, {emailid}, {listingurl}, {advertiserurl}",
                            i, searchRequest.Query, templateVars.EmailId, listing.ListingUrl, listing.AdvertiserUrl);
                    }

                    clickUrls["Listing" + listing.ListingIndex + "ClickUrl"] = listing.ClickUrl;
                }

                // TODO: Eventually merge this with other ClickUrlStorage from TaboolaListings and IndeedListings
                if (storeClickURLs)
                {
                    ClickUrlStorage.SaveClickUrlsToDynamo(templateVars.EmailId, clickUrls);
                }
            }

            if (shouldSearchIndeed)
            {
                var indeedResultCount = 10;
                var indeedSearchResult = Providers.IndeedProvider.GetListings(searchRequest.Query, searchRequest.Ip, indeedResultCount);
                indeedSearchResult.ClickUrlFunc = listing => Tracker.WrapPlainUrl(clickWrapper, listing.ClickUrl, null);

                var indeedListings = indeedSearchResult.Listings.ToList();

                templateVars.IndeedListings = new List<IndeedListingTemplateVariables>();
                for (var i = 0; i < indeedResultCount; i++)
                {
                    var indeedListing = new IndeedListingTemplateVariables();
                    indeedListing.ListingIndex = i + 1;
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex, indeedListing);
                    templateVars.IndeedListings.Add(indeedListing);
                    if (i >= indeedListings.Count)
                    {
                        continue;
                    }

                    indeedListing.Rank = indeedListings[i].Rank;
                    indeedListing.Title = indeedListings[i].Title;
                    indeedListing.Advertiser = indeedListings[i].SiteHost;
                    indeedListing.Description = indeedListings[i].Description;
                    indeedListing.AdvertiserUrl = indeedListings[i].GetAdvertiserUrl();
                    indeedListing.ListingUrl = indeedListings[i].ClickUrl;
                    indeedListing.ClickUrl = indeedSearchResult.ClickUrlFunc(indeedListings[i]);

                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "Title", indeedListing.Title);
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "Advertiser", indeedListing.Advertiser);
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "Description", indeedListing.Description);
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "AdvertiserUrl", indeedListing.AdvertiserUrl);
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "ListingUrl", indeedListing.ListingUrl);
                    TrySetProperty(templateVars, "IndeedListing" + indeedListing.ListingIndex + "ClickUrl", indeedListing.ClickUrl);
                }
            }

            // NEW: If Taboola Template, ping TaboolaProvider and add Taboola Listings to the templateVars
            // shouldSearchTaboola = true;
            if (shouldSearchTaboola)
            {
                var taboolaResultCount = 10;

                // TODO: Change searchRequest.ServeUrl to appropriate Source URL of the request
                // var taboolaSearchResult = Providers.TaboolaProvider.GetListings(searchRequest.Ip, taboolaResultCount, searchRequest.ServeUrl, searchRequest.Query, searchRequest.Query);
                // taboolaSearchResult.ClickUrlFunc = listing => Tracker.WrapPlainUrl(clickWrapper, listing.ClickUrl, null);


                SearchResult taboolaSearchResult;
                if (shouldSearchTaboola)
                {
                    // AK: look into fixing ClickUrl 
                    var sw = System.Diagnostics.Stopwatch.StartNew();
                    taboolaSearchResult = Providers.TaboolaProvider.GetListings(searchRequest.Ip, taboolaResultCount, searchRequest.ServeUrl, searchRequest.Query, searchRequest.Query);
                    RedisHelper.AddRequestTime("TaboolaSearch", sw.Elapsed.TotalMilliseconds);
                    taboolaSearchResult.ClickUrlFunc = listing => Tracker.WrapPlainUrl(clickWrapper, listing.ClickUrl, null);
                    // taboolaSearchResult.ClickUrlFunc = listing => Tracker.MakeTrackerUrl(searchRequest.ServeUrl,
                    //    clickWrapper, requestDict, listing, taboolaSearchResult);
                    // taboolaSearchResult.ClickUrlFunc = listing => Tracker.MakeTaboolaTrackerUrl(searchRequest.ServeUrl,
                    //    clickWrapper, requestDict, listing, taboolaSearchResult);
                }
                else
                {
                    taboolaSearchResult = new SearchResult();
                }

                var taboolaListings = taboolaSearchResult.Listings.ToList();

                var taboolaClickUrls = new Dictionary<string, string>();
                templateVars.TaboolaListings = new List<TaboolaListingTemplateVariables>();
                for (var i = 0; i < taboolaResultCount; i++)
                {
                    var taboolaListing = new TaboolaListingTemplateVariables();
                    taboolaListing.ListingIndex = i + 1;
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex, taboolaListing);
                    templateVars.TaboolaListings.Add(taboolaListing);
                    if (i >= taboolaListings.Count)
                    {
                        continue;
                    }

                    taboolaListing.Rank = taboolaListings[i].Rank;
                    taboolaListing.Title = taboolaListings[i].Title;
                    taboolaListing.Advertiser = taboolaListings[i].SiteHost;
                    taboolaListing.Description = taboolaListings[i].Description;
                    taboolaListing.AdvertiserUrl = taboolaListings[i].GetAdvertiserUrl();
                    taboolaListing.ListingUrl = taboolaListings[i].ClickUrl;
                    taboolaListing.ClickUrl = taboolaSearchResult.ClickUrlFunc(taboolaListings[i]);
                    // taboolaListing.ClickUrl = searchResult.ClickUrlFunc(taboolaListings[i]);
                    // taboolaListing.ThumbnailImageUrl = taboolaListings[i].ThumbnailImageUrl;
                    taboolaListing.ThumbnailImageUrl = HttpUtility.UrlDecode(taboolaListings[i].ThumbnailImageUrl.Substring(97));

                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "Title", taboolaListing.Title);
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "Advertiser", taboolaListing.Advertiser);
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "Description", taboolaListing.Description);
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "AdvertiserUrl", taboolaListing.AdvertiserUrl);
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "ListingUrl", taboolaListing.ThumbnailImageUrl);  // Stores Thumbnail Image URL
                    TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "ClickUrl", taboolaListing.ClickUrl);
                    // TrySetProperty(templateVars, "TaboolaListing" + taboolaListing.ListingIndex + "ThumbnailImageUrl", taboolaListing.ThumbnailImageUrl); // Thumbnail Image URL gets stored into ListingUrl instead

                    taboolaClickUrls["TaboolaListing" + taboolaListing.ListingIndex + "ClickUrl"] = taboolaListing.ClickUrl;
                }

                // TODO: Eventually merge this with other ClickUrlStorage from Listings and IndeedListings
                if (storeClickURLs)
                {
                    ClickUrlStorage.SaveClickUrlsToDynamo(templateVars.EmailId, taboolaClickUrls);
                }
            }

            var renderedTemplate = Nustache.Core.Render.StringToString(html, templateVars);

            if (Request["mimeType"] != null)
            {
                if (Request["mimeType"] == "text/css")
                {
                    renderedTemplate = renderedTemplate.Replace("&#34;", "\\\"");
                    renderedTemplate = WebUtility.HtmlDecode(renderedTemplate);
                }
                return Content(renderedTemplate, Request["mimeType"]);
            }
            else
            {
                return Content(renderedTemplate);
            }
        }
        
        public ActionResult DynamicClick(string emailId, string listing, string siteLink = null)
        {
            string clickUrlKey;
            if (!String.IsNullOrEmpty(siteLink))
            {
                if (siteLink == "TaboolaListing")
                {
                    clickUrlKey = "TaboolaListing" + listing + "ClickUrl";
                }
                else
                {
                    clickUrlKey = "Listing" + listing + "SiteLink" + siteLink + "ClickUrl";
                }
            }
            else 
            {
                clickUrlKey = "Listing" + listing + "ClickUrl";
            }

            Uri mobsearchClickUri = null;

            var wrappedClickUrl = ClickUrlStorage.TryGetClickUrlFromDynamo(emailId, clickUrlKey);
            if (String.IsNullOrEmpty(wrappedClickUrl))
            {
                // invalid click
                Log.Error("Invalid Click. Wrapped Click URL was empty with emailId {emailid}, listing {listing}, siteLink {siteLink}.", emailId, listing, siteLink);
                return RedirectToAction("Index", "Home");
            }
            Uri wrappedClickUri = new Uri(wrappedClickUrl);
            var mobsearchClickUrl = HttpUtility.ParseQueryString(wrappedClickUri.Query).Get("url");
            if (mobsearchClickUrl == null)
            {
                mobsearchClickUri = wrappedClickUri;
            }
            else
            {
                mobsearchClickUri = new Uri(mobsearchClickUrl);
            }

            var dictionary = HttpUtility.ParseQueryString(mobsearchClickUri.Query).ToDictionary();
            dictionary.AddOrUpdate("ip", RequestHelper.GetIp(HttpContext));
            dictionary.AddOrUpdate("userAgent", Request.UserAgent);
            dictionary.AddOrUpdate("timestamp", DateTime.UtcNow.ToString("O"));
            
            if (siteLink == "TaboolaListing")
            {
                // TODO: Probably want to adjust the API Taboola URL around here to add user ID parameters and such
                HostingEnvironment.QueueBackgroundWorkItem(ct => RelayMessage(dictionary));
                HostingEnvironment.QueueBackgroundWorkItem(ct =>
                {
                    Metrics.Time(() => Tracker.TrackClick(dictionary), "trackClick");
                });

                var tabModel = new FormSubmitModel();
                tabModel.U = mobsearchClickUrl;
                return View("FormSubmit", tabModel);
            }

            var listingUrl = string.Empty;
            dictionary.TryGetValue("u", out listingUrl);
            if (String.IsNullOrEmpty(listingUrl))
            {
                Log.Error("Listing URL is empty",emailId, listing);
                return RedirectToAction("Index", "Home");
            }

            HostingEnvironment.QueueBackgroundWorkItem(ct => RelayMessage(dictionary));
            HostingEnvironment.QueueBackgroundWorkItem(ct =>
            {
                Metrics.Time(() => Tracker.TrackClick(dictionary), "trackClick");
            });

            var model = new FormSubmitModel();
            model.U = listingUrl;
            return View("FormSubmit", model);
        }

        /// <summary>
        /// Keep track if this was the first or 2nd click. 
        /// This works by finding the event_subtype on the clickwrapper,
        /// if it matches the pattern click[clickNumber] then clickNumber is incremented,
        /// otherwise we add click1 as the event_subtype parameter.
        /// 
        /// This works when one template links to another and there's no place to change the 
        /// click wrapper, it happens automatically
        /// </summary>
        private string IncrementClickCountOnClickWrapper(string clickWrapper, string baseValue = "click")
        {
            if (clickWrapper == null)
            {
                return null;
            }
            var pattern = "event_subtype=" + baseValue + @"([\d+])";
            var matchSuccess = false;
            clickWrapper = Regex.Replace(clickWrapper, pattern, m =>
            {
                matchSuccess = true;
                var clickIndex = int.Parse(m.Groups[1].Value);
                return "event_subtype=" + baseValue + (clickIndex + 1);
            });
            if (!matchSuccess)
            {
                clickWrapper = clickWrapper.AddQueryParam("event_subtype", baseValue + "1");
            }
            return clickWrapper;
        }

        private string StrongWords(string title, string query)
        {
            var words = query.Split(new[] { " " }, StringSplitOptions.RemoveEmptyEntries);
            foreach (var word in words)
            {
                title = Regex.Replace(title, word, m => string.Format("<strong>" + m.Value + "</strong>"), RegexOptions.IgnoreCase);
            }
            return title;
        }

        public bool TrySetProperty(object o, string propertyName, object value)
        {
            var didSet = false;
            var property = o.GetType().GetProperty(propertyName);
            if (property != null)
            {
                property.SetValue(o, value);
                didSet = true;
            }
            return didSet;
        }

        // GET : /Test
        // Just a page for viewing different options
        public ActionResult Test()
        {
            return View();
        }

        public static List<string> GetAdvertiserBlacklistForKeyword(String keyword)
        {
            var _dynamoClient = new Dynamo(new BasicAWSCredentials(ConfigUtilities.GetAppSetting("AWSAccessKey"), ConfigUtilities.GetAppSetting("AWSSecretKey")), RegionEndpoint.USWest2);

            var jsonList = _dynamoClient.GetAll("prd_2click_blacklist2");
            var blacklist = jsonList.Where(x =>
            {
                var jObj = JObject.Parse(x);
                return (string)jObj["keyword"] == keyword;
            }).Select(x =>
            {
                var jObj = JObject.Parse(x);
                return (string)jObj["advertiserUrl"];
            }).ToList();
            return blacklist;
        }
    }

    public class MultiTermsViewModel 
    {
        public string ActiveKeyword { get; set; }
        public string Subid { get; set; }
        public string ClickWrapper { get; set; }
        public string FooterHtml { get; set; }
    }
}
 