﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Net;
using System.Web;
using System.Web.Mvc;
using System.Web.Optimization;
using System.Web.Routing;
using Mobile.Search.Web.Helpers;
using Serilog;
using Serilog.Events;
using StatsdClient;

namespace Mobile.Search.Web
{
    public class MvcApplication : System.Web.HttpApplication
    {
        public static string InstanceId;
        private static void ConfigureLogger()
        {
            InstanceId = GetInstanceId();
            Log.Logger = new LoggerConfiguration()
                .MinimumLevel.Verbose()
                //Sinks
                .WriteTo.RollingFile(ConfigurationManager.AppSettings.Get("LogPath"))
                .WriteTo.Seq("http://ec2-52-90-180-25.compute-1.amazonaws.com:5341", LogEventLevel.Verbose)
                //Enrichers
                .Enrich.WithProperty("AppName", "MobSearch Web")
#if DEBUG
                .Enrich.WithProperty("Env", "dev")
#else
                .Enrich.WithProperty("Env", "prd")
#endif
                .Enrich.WithProperty("InstanceId", InstanceId)
                .Enrich.WithEnvironmentUserName()
                .Enrich.WithMachineName()
                .CreateLogger();
        }
        protected void Application_Start()
        {
            AreaRegistration.RegisterAllAreas();
            FilterConfig.RegisterGlobalFilters(GlobalFilters.Filters);
            RouteConfig.RegisterRoutes(RouteTable.Routes);
            BundleConfig.RegisterBundles(BundleTable.Bundles);
            ConfigureLogger();
            JobScheduler.Start();
            var prefix = "search." + Environment.MachineName + "." + GetInstanceId();

            var metricsConfig = new MetricsConfig
            {
                StatsdServerName = ConfigurationManager.AppSettings["StatsDServer"],
                Prefix = prefix
            };

            Metrics.Configure(metricsConfig);
            Log.Information("Application Started");

        }

        protected void Application_BeginRequest()
        {
            HttpContext.Current.Items["Stopwatch"] = System.Diagnostics.Stopwatch.StartNew();
        }

        protected void Application_EndRequest()
        {
            try
            {
                if (InstanceId == "NotEc2")
                    return;
                var endpoint = HttpContext.Current.Request.Url.AbsolutePath;
                var sw = (System.Diagnostics.Stopwatch) HttpContext.Current.Items["Stopwatch"];
                if (sw == null) return;

                sw.Stop();
                RedisHelper.AddRequestTime(endpoint, sw.Elapsed.TotalMilliseconds);
            }
            catch (Exception x)
            {
                Log.Error(x, "Error loading request time to redis");
            }
        }
        static string GetInstanceId()
        {
            const string endpoint = "http://169.254.169.254/latest/meta-data/instance-id";
            var client = new TimedWebClient() { Timeout = 500};
            try
            {
                return client.DownloadString(endpoint);
            }
            catch
            {
                return "NotEc2";
            }
        }

        protected void Application_Error(Object sender, EventArgs e)
        {
            var exception = Server.GetLastError().GetBaseException();
            if (exception is HttpException)
            {
                var httpEx = exception as HttpException;
                try
                {
                    if (httpEx?.GetHttpCode() == 404)
                    {
                        return;
                    }
                }
                catch
                {
                    Log.Error(exception, "Uncaught exception {ExceptionType}", exception.GetType().Name);
                }

                try
                {
                    Log.Error(httpEx, "Unhandled HTTP Exception. Path={Path} QueryString={QueryString}", Request.Path,
                        Request.QueryString.AllKeys.Distinct().ToDictionary(x => x, x => Request.QueryString[x]));
                }
                catch (Exception x)
                {
                    Log.Error(x, "Error logging exception {ex}", httpEx.Message);
                }
            }
            else
            {
                Log.Error(exception, "Uncaught exception {ExceptionType}", exception.GetType().Name);
            }
        }
    }
    public class TimedWebClient : WebClient
    {
        // Timeout in milliseconds, default = 600,000 msec
        public int Timeout { get; set; }

        public TimedWebClient()
        {
            this.Timeout = 600000;
        }

        protected override WebRequest GetWebRequest(Uri address)
        {
            var objWebRequest = base.GetWebRequest(address);
            objWebRequest.Timeout = this.Timeout;
            return objWebRequest;
        }
    }
}
