﻿using System;
using System.Collections.Generic;
using System.Linq;
using Amazon;
using Amazon.CloudWatch;
using Amazon.CloudWatch.Model;
using Amazon.Runtime;
using Quartz;
using Serilog;

namespace Mobile.Search.Web.Helpers
{
    [DisallowConcurrentExecution]
    public class PublishCloudWatchMetricsJob : IJob
    {
        private readonly AmazonCloudWatchClient _cloudWatchClient;
        public PublishCloudWatchMetricsJob()
        {
            _cloudWatchClient = new AmazonCloudWatchClient(new EnvironmentAWSCredentials(), RegionEndpoint.USWest2);
        }

        private const string RedisLockKey = "PublishCloudWatchMetricsLock";
        public void Execute(IJobExecutionContext context)
        {
            var lastPush = RedisHelper.GetLastRequestTimesPush();
            if (lastPush == null)
            {
                RedisHelper.SetLastRequestTimesPush();
                return;
            }

            if (DateTime.UtcNow - TimeSpan.FromSeconds(45) > lastPush.Value)
            {
                var lockToken = Guid.NewGuid().ToString();
                try
                {
                    if (!RedisHelper.AcquireLock("PublishCloudWatchMetricsLock", lockToken, TimeSpan.FromSeconds(10)))
                        return;

                    var data = RedisHelper.GetAndRemoveRequestTimeAverages();

                    if (!data.Any())
                        return;

                    var metrics = new List<MetricDatum>();
                    foreach (var kvp in data)
                    {
                        double avgTime = kvp.Value;
                        metrics.Add(new MetricDatum
                        {
                            Dimensions = new List<Dimension>
                            {
                                new Dimension
                                {
                                    Name = "Mobsearch Endpoint",
                                    Value = kvp.Key
                                }
                            },
                            MetricName = "Request Time",
                            StatisticValues = new StatisticSet(),
                            Timestamp = DateTime.Now,
                            Unit = StandardUnit.Milliseconds,
                            Value = avgTime
                        });
                    }


                    var request = new PutMetricDataRequest
                    {
                        MetricData = metrics,
                        Namespace = "MobSearch Metrics"
                    };


                    try
                    {
                        _cloudWatchClient.PutMetricData(request);
                        Log.Verbose("Published {MetricCount} Cloud Watch Metrics", data.Count);
                    }
                    catch (Exception x)
                    {
                        Log.Error(x, "Error pushing cloud watch metrics");
                    }
                    RedisHelper.SetLastRequestTimesPush();
                }
                finally
                {
                    RedisHelper.ReleaseLock(RedisLockKey, lockToken);
                }
            }
        }
    }
}