﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using Serilog;
using StackExchange.Redis;

namespace Mobile.Search.Web.Helpers
{
    public class RedisHelper
    {
        private static readonly ILogger Log = Serilog.Log.ForContext<RedisHelper>(); 
        private static readonly string RedisConnStr = ConfigurationManager.AppSettings["RedisConnString"];

        private static readonly Lazy<ConnectionMultiplexer> LazyConnection =
            new Lazy<ConnectionMultiplexer>(() => ConnectionMultiplexer.Connect(RedisConnStr));

        public static ConnectionMultiplexer Multiplexer
        {
            get { return LazyConnection.Value; }
        }

        public static void AddLastKeywordClicked(string uuid, string keyword)
        {
            if (string.IsNullOrEmpty(uuid) || string.IsNullOrEmpty(keyword))
            {
                return;
            }
            var db = Multiplexer.GetDatabase(0);
            db.HashSet("user:" + uuid, "last_clicked_keyword", keyword);
            db.HashSet("user:" + uuid, "last_clicked_time", DateTime.UtcNow.ToString(""));
        }

        public static void AddLastKeywordOpened(string uuid, string keyword)
        {
            if (string.IsNullOrEmpty(uuid) || string.IsNullOrEmpty(keyword))
            {
                return;
            }
            var db = Multiplexer.GetDatabase(0);
            db.HashSet("user:" + uuid, "last_opened_keyword", keyword);
            db.HashSet("user:" + uuid, "last_opened_time", DateTime.UtcNow.ToString(""));
        }

        public static Dictionary<string,string> GetLastKeywords(string uuid)
        {
            var db = Multiplexer.GetDatabase(0);
            return db.HashGetAll("user:" + uuid)
                .ToDictionary()
                .ToDictionary(x => x.Key.ToString(), x => x.Value.ToString());
        }

        private static string RequestTimesEndpointKey = "MobSearch:RequestTimes:Endpoints";
        private static string RequestTimesLastPushTimeKey = "MobSearch:RequestTimes:LastPush";
        private static string RequestTimesFmtKey = "MobSearch:RequestTimes:{0}";
        public static void AddRequestTime(string endpoint, double value)
        {
            try
            {
                var db = Multiplexer.GetDatabase(0);
                db.SetAdd(RequestTimesEndpointKey, endpoint.ToLower());
                db.ListLeftPush(string.Format(RequestTimesFmtKey, endpoint.ToLower()), value);
            }
            catch (Exception x)
            {
                Log.Error(x, "Error adding request time to {endpoint} with value {value}", endpoint, value);
            }

        }

        public static Dictionary<string, double> GetAndRemoveRequestTimeAverages()
        {
            var dict = new Dictionary<string, double>();
            var db = Multiplexer.GetDatabase(0);
            var members = db.SetMembers(RequestTimesEndpointKey).Select(x => x.ToString());

            foreach (var endpoint in members)
            {
                var items = db.ListRange(string.Format(RequestTimesFmtKey, endpoint));
                if (items.Any())
                {
                    dict.Add(endpoint, items.Average(x => double.Parse(x)));
                }
                db.KeyDelete(string.Format(RequestTimesFmtKey, endpoint));
            }

            db.KeyDelete(RequestTimesEndpointKey);
            return dict;
        }

        public static DateTime? GetLastRequestTimesPush()
        {
            var db = Multiplexer.GetDatabase(0);
            var data = db.StringGet(RequestTimesLastPushTimeKey);

            if (!data.IsNullOrEmpty)
            {
                var dt = default(DateTime);
                if (DateTime.TryParse(data.ToString(), out dt))
                {
                    return dt;
                }
            }
            return null;
        }

        public static bool AcquireLock(string name, string token, TimeSpan expiry)
        {
            var db = Multiplexer.GetDatabase(0);
            return db.LockTake(name, token, expiry);
        }

        public static void ReleaseLock(string name, string token)
        {
            var db = Multiplexer.GetDatabase(0);
            db.LockRelease(name, token);
        }
        public static void SetLastRequestTimesPush()
        {
            var db = Multiplexer.GetDatabase(0);
            db.StringSet(RequestTimesLastPushTimeKey, DateTime.UtcNow.ToString("O"));
        }

    }
}